#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

__all__ = ["Extractor"]

import jax.numpy as jnp
import nlopt
import numpy as np
from jax import grad, jit
from jax.config import config
from scipy.optimize import minimize

from coaxtract.constants import *
from coaxtract.stack import Stack


def mse(a, b):
    return jnp.mean(jnp.abs(a - b) ** 2)


def eps2opt(eps):
    return np.hstack([eps.real, eps.imag])


class Extractor:
    def __init__(
        self,
        setup,
        frequencies,
        Rmeas,
        scale=True,
        guess="random",
        options=None,
        eps_min=-1e9 * (1 + 1j),
        eps_max=1e9 * (1 + 1j),
        single_layer=False,
        compile=False,
        verbose=False,
        n_int=700,
        alpha=0.5,
        model="constant",
        bounds=None,
        constraints=False,
        backend="scipy",
    ):
        self.setup = setup
        self.constraints = constraints
        self.bounds = bounds
        self.backend = backend
        if backend not in ["scipy", "nlopt"]:
            raise ValueError("backend must be scipy or nlopt")
        if model not in ["constant", "wideband", "cole-cole"]:
            raise ValueError("model must be wideband, constant or cole-cole")
        self.film = None
        for i, lay in enumerate(setup.stack.layers):
            if lay.name == "unknown":
                self.film = lay
                self.film_index = i
        if self.film is None:
            raise ValueError("layer named unknown must be contained in the setup stack")

        self.frequencies = frequencies
        self.Rmeas = Rmeas
        self.probe = setup.probe
        self.guess = guess
        self.nfreq = len(Rmeas)

        self.model = model
        if self.model == "constant":
            self.nvar = 2
        elif self.model == "wideband":
            self.nvar = 2 * self.nfreq
        else:
            self.nvar = 4
        self.verbose = verbose
        self.scale = scale
        self.single_layer = single_layer
        self.compile = compile
        self.alpha = alpha

        self.n_int = n_int
        default_options = {
            "disp": int(self.verbose),
            "maxcor": 250,
            "ftol": 1e-16,
            "gtol": 1e-16,
            "eps": 1e-12,
            # "eps": 1e-11,
            "maxfun": 15000,
            "maxiter": 1300,
            "iprint": 1,
            "maxls": 200,
            "finite_diff_rel_step": None,
        }

        self.options = options or default_options
        self.options["disp"] = int(self.verbose)

        self.eps_min = eps_min
        self.eps_max = eps_max

        if self.compile:
            self.fun_jit = jit(self.fun)
            self.grad_fun_jit = jit(grad(self.fun_jit))
        else:
            self.fun_jit = self.fun
            self.grad_fun_jit = grad(self.fun_jit)

    def thin_film_approx(self):
        Zc = self.probe.impedance
        Zmeas = Zc * (1 + self.Rmeas) / (1 - self.Rmeas)
        out = 1 / (
            1j * 2 * pi * self.frequencies * self.film.thickness * Zmeas * epsilon_0
        )
        # out[np.isnan(out)] = 0
        return out

    def opt2eps(self, x):
        if self.model == "cole-cole":
            omega = 2 * pi * self.frequencies * 1e-9
            return cole_cole(omega, *x) - 1e-6
        else:
            eps_re, eps_im = x[: int(self.nvar / 2)], x[int(self.nvar / 2) :]
            return eps_re + 1j * eps_im

    def reflection(self, eps_film):
        self.setup.stack.layers[self.film_index].epsilon = eps_film
        stack = self.setup.stack
        layers = stack.layers
        layers[self.film_index].epsilon = eps_film
        self.setup.stack = Stack(layers, stack.termination)
        r = self.setup.reflection(
            self.frequencies, n_int=self.n_int, alpha=self.alpha, iopt=self.film_index
        )
        return jnp.array(r)

    def fun(self, x):
        eps_film = self.opt2eps(x) * jnp.ones(self.nfreq)
        R_ana = self.reflection(eps_film)
        obj = mse(self.Rmeas, R_ana)
        if self.scale:
            obj /= self.film.thickness
        return obj

    def init_bounds(self):
        if self.bounds is not None:
            return np.array(self.bounds)
        else:
            if self.model == "cole-cole":
                bounds = [(self.eps_min.real, self.eps_max.real)]
                bounds += [(self.eps_min.real, self.eps_max.real)]
                bounds += [(-1e6, -1e-16)]
                bounds += [(0, 1)]
            else:
                bounds = [
                    (self.eps_min.real, self.eps_max.real)
                    for i in range(int(self.nvar / 2))
                ]
                bounds += [
                    (self.eps_min.imag, self.eps_max.imag)
                    for i in range(int(self.nvar / 2))
                ]
            return np.array(bounds)

    def init_guess(self):
        re = np.array(self.thin_film_approx().real)
        im = np.array(self.thin_film_approx().imag)
        eps_approx = re.astype(float) + 1j * im.astype(float)

        if self.guess == "random":
            eps_random = self.eps_min + (self.eps_max - self.eps_min) * np.random.rand(
                self.nvar
            )
            return eps_random
        elif self.guess == "approx":
            return eps2opt((eps_approx))
        elif self.guess == "approx mean":
            guess = eps2opt((np.mean(eps_approx) * np.ones(self.nfreq)))
            return guess
        elif self.guess == "approx low freq":
            guess = eps2opt((eps_approx[0] * np.ones(self.nfreq)))
            return guess
        elif not isinstance(self.guess, str):
            if np.isscalar(self.guess):
                guess = self.guess
                if self.model == "constant":
                    return eps2opt(guess)
                else:
                    initial_guess = np.complex64(guess * np.ones(self.nfreq))
                    return eps2opt(initial_guess)
            elif len(self.guess) == self.nfreq:
                guess = self.guess
                return eps2opt(np.complex64(guess))
            elif self.model == "cole-cole":
                return self.guess
            else:
                raise ValueError("Wrong initial guess")

    def run(self):
        bounds = self.init_bounds()
        initial_guess = self.init_guess()
        if self.backend == "scipy":
            opt = minimize(
                self.fun_jit,
                initial_guess,
                bounds=bounds,
                tol=1e-12,
                options=self.options,
                jac=self.jacobian,
                method="L-BFGS-B",
            )
        else:

            def fun_nlopt(x, gradn):
                y = self.fun_jit(x)
                y = np.float64(y)
                # y = np.array(y.astype(float))
                if gradn.size > 0:
                    dy = np.float64(self.jacobian(x))
                    # dy = dy.astype(float)
                    gradn[:] = dy
                if self.verbose:
                    print(f">>> objective = {y}")
                    eps = self.opt2eps(x)
                    print(f"    mean permittivity = {np.mean(eps)}")
                    print(f"    std permittivity  = {np.std(eps)}")
                return y

            opt = nlopt.opt(nlopt.LD_MMA, self.nvar)
            bounds = self.init_bounds()
            opt.set_lower_bounds(bounds[:, 0])
            opt.set_upper_bounds(bounds[:, 1])
            opt.set_maxeval(self.options["maxiter"])
            opt.set_ftol_rel(self.options["ftol"])
            opt.set_xtol_rel(self.options["eps"])
            opt.set_min_objective(fun_nlopt)

            if self.model == "cole-cole" and self.constraints:

                def _fc(x, i):
                    eps = self.opt2eps(x)
                    c1 = -eps.real + self.eps_min.real
                    c2 = eps.imag - self.eps_max.imag
                    return jnp.hstack([c1, c2])[i]

                _grad_fc = jit(grad(_fc))

                def fc(result, x, gradn):
                    if gradn.size > 0:
                        gradn[:] = [_grad_fc(x, i) for i in range(2 * self.nfreq)]

                    result[:] = [_fc(x, i) for i in range(2 * self.nfreq)]
                    return result

                tols = np.ones(2 * self.nfreq) * 1e-6
                opt.add_inequality_mconstraint(fc, tols)

            xopt = opt.optimize(initial_guess)
            fopt = opt.last_optimum_value()
            opt.x = xopt
            opt.fun = fopt

        eps_opt = self.opt2eps(opt.x) * jnp.ones(self.nfreq)
        self.eps_opt = eps_opt
        self.opt = opt
        return opt

    def jacobian(self, x):
        out = np.array(self.grad_fun_jit(x))
        out = out.astype(float)
        return out


#### permittivity models


def cole_cole(omega, eps_inf, eps_static, tau, alpha):
    return eps_inf + (eps_static - eps_inf) / (1 + (1j * omega * tau) ** (1 - alpha))


if __name__ == "__main__":

    import coaxtract as cx

    probe = cx.CoaxProbe()
    freqs = np.linspace(1, 2, 10) * 1e9
    Rmeas_cal = np.ones_like(freqs)
    t_film = 0.5e-3
    layopt = cx.Layer(1, t_film, name="unknown")
    opt = cx.Stack([layopt])
    exp = cx.Setup(opt, probe)

    for l in exp.stack.layers:
        if l.name == "opt":
            film = l

    self = cx.Extractor(
        exp,
        freqs,
        Rmeas_cal,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


from jax.config import config

config.update("jax_enable_x64", True)
# config.update("jax_debug_nans", True)

from .__about__ import __author__, __description__, __version__
from .calibration import *
from .constants import *
from .helpers import *
from .io import *
from .opt import *
from .probe import *
from .stack import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = [
    "effective_impedance",
    "smooth",
    "meas2complex",
    "complex2aphi",
    "truncate",
    "Timer",
]

import time

import numpy as np
import scipy.signal as signal

from .constants import *


def effective_impedance(Z, epsilon, thickness, freq):
    k0 = 2 * pi * freq / c
    n = epsilon ** 0.5
    T = np.tan(k0 * n * thickness)
    z = Z0 / n
    return z * (Z + 1j * z * T) / (z + 1j * Z * T)


def smooth(data, cutoff=0.05, order=3):
    """Data smoothing.

    Parameters
    ----------
    data : array
        The data to smooth.
    cutoff : float
        Filter cutoff frequency (the default is 0.05).
    order : int
        Filter order (the default is 3).

    Returns
    -------
    smooth_data: array
        Smoothed data.

    """
    # Buterworth filter
    B, A = signal.butter(order, cutoff, output="ba")
    smooth_data = signal.filtfilt(B, A, data)
    return smooth_data


def meas2complex(A, phi):
    return 10 ** (A / 20) * np.exp(1j * phi * pi / 180)


def complex2aphi(z):
    return np.abs(z), np.unwrap(np.angle(z)) * 180 / pi


def truncate(freqs, min_freq, max_freq, dspl=0):
    return np.where(np.logical_and(freqs >= min_freq, freqs <= max_freq))[0][::dspl]


class TimerError(Exception):

    """A custom exception used to report errors in use of Timer class"""


class Timer:
    def __init__(self):

        self._start_time = None

    def start(self):
        """Start a new timer"""

        if self._start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self):
        """Stop the timer, and report the elapsed time"""

        if self._start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        print(f"Elapsed time: {elapsed_time:0.4f} seconds")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = ["Calibration"]

import warnings

import jax.numpy as jnp

from .constants import *
from .special import *
from .stack import *

# def j0(z):
#     return j0_approx(z, N=100)


# warnings.filterwarnings("ignore")


class Calibration:
    def __init__(
        self,
        probe,
        epsilon_load=4 * (1 - 0.01j),
        thickness_load=1.3e-3,
        epsilon_open=1 - 0j,
        air_gap=0,
        bessel="jax",
    ):

        self.cases = ["open", "short", "load"]
        self.epsilon_load = epsilon_load
        self.thickness_load = thickness_load

        self.air_gap = air_gap
        self.epsilon_open = epsilon_open
        load_layer = Layer(epsilon_load, thickness_load)
        if self.air_gap == 0:
            load = Stack([load_layer], termination="PEC")
        else:
            air_layer = Layer(epsilon_open, air_gap)
            load = Stack([air_layer, load_layer], termination="PEC")
        self.load_setup = Setup(
            load,
            probe,
            bessel=bessel,
        )

        short_layer = Layer(epsilon_open, air_gap)
        short = Stack([short_layer], termination="PEC")
        self.short_setup = Setup(
            short,
            probe,
            bessel=bessel,
        )

        open_layer = Layer(epsilon=epsilon_open, thickness=None)
        open = Stack([open_layer], termination="infinite")
        self.open_setup = Setup(
            open,
            probe,
            bessel=bessel,
        )

        self.probe = probe
        self.bessel = bessel

        # self.S11 = dict(theory=None, measured=None)

    def reflection(self, freqs, case, **kwargs):
        if case not in self.cases:
            raise ValueError(f"case must be in {self.cases}")

        if case == "short":
            if self.air_gap == 0:
                return -jnp.ones_like(freqs)
            else:
                return self.short_setup.reflection(freqs, **kwargs)

        elif case == "load":
            iopt = 0 if self.air_gap == 0 else 1
            kwargs["iopt"] = iopt
            return self.load_setup.reflection(freqs, **kwargs)
        else:
            return self.open_setup.reflection(freqs, **kwargs)

    def get_theoretical_reflection(self, freqs, **kwargs):
        R = {}
        for case in self.cases:
            R[case] = self.reflection(freqs, case, **kwargs)
        return R

    def run(self, freqs, Rmeasured, **kwargs):
        nfreq = len(freqs)
        Id = jnp.eye(nfreq)
        R = self.get_theoretical_reflection(freqs, **kwargs)
        b = [
            [jnp.diag(R[case]), Id, -jnp.diag(R[case] * Rmeasured[case])]
            for case in self.cases
        ]
        B = jnp.block(b)
        v = jnp.hstack(list(Rmeasured.values()))
        sol = jnp.linalg.inv(B) @ v
        a, b, c = sol[:nfreq], sol[nfreq : 2 * nfreq], sol[-nfreq:]
        self.coefficients = dict(a=a, b=b, c=c)
        return self.coefficients

    def apply(self, Rmeas, freqspan=None):
        if freqspan is None:
            freqspan = range(len(Rmeas))
        return (Rmeas - self.coefficients["b"][freqspan]) / (
            -self.coefficients["c"][freqspan] * Rmeas + self.coefficients["a"][freqspan]
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


"""
From :cite:p:`Zhou2017`.
"""
# Zhou, L., Ju, Y., Wang, P. & Pei,  and Y.
# Fast and Stable Integration Method for the Aperture Admittance of an
# Open-Ended Coaxial Probe Terminated into Low-Loss Dielectrics.
# Progress In Electromagnetics Research M 55, 211–219 (2017)

from jax.config import config

# config.update("jax_debug_nans", True)
config.update("jax_debug_nans", False)

import jax.numpy as jnp
import numpy as np
import scipy.integrate as si
from jax import ops

from .constants import *


def path_integral(
    integrand,
    eps,
    path="complex",
    alpha=0.5,
    int_method="trapz",
    n_int=100,
    divmax=5,
):

    if int_method not in ["trapz", "romberg", "romberg_jax"]:
        raise ValueError('int_method must be one of "trapz", "romberg", "romberg_jax"')

    if path not in ["complex", "real"]:
        raise ValueError('path must be one of "complex", "real"')

    _A0 = (eps.real) ** 0.5
    if jnp.isscalar(_A0):
        n = 1
    else:
        n = len(_A0)

    def broadcast(t):
        if int_method == "trapz":
            A0 = jnp.broadcast_to(_A0, (len(t), n))
            t1 = jnp.broadcast_to(t, (n, len(t))).T
            return A0, t1
        else:
            return _A0, t

    def y1(t):

        A0, t = broadcast(t)
        return 2 * A0 * integrand(2 * A0 * t)

    def y2(t):
        A0, t = broadcast(t)
        if jnp.isscalar(t) and t == 0:
            return 0
        else:
            epstol = 1e-14
            result = jnp.where(
                t == 0.0,
                0.0,
                2 * A0 / (t + epstol) ** 2 * integrand(2 * A0 / (t + epstol)),
            )

            # return result
            # return jnp.where(
            #     t == 0.0, 0.0, 2 * s0 / (t + 1e-16) ** 2 * f(2 * s0 / (t + 1e-16))
            # )
            return jnp.where(jnp.isfinite(result), result, 0)

    ### integration along an elliptical path in the complex plane
    def elliptical_path(theta, alpha=alpha):
        A0, theta = broadcast(theta)
        return A0 + A0 * (jnp.cos(theta) + 1j * alpha * jnp.sin(theta))

    def yc(z, alpha=alpha):
        A0, t = broadcast(z)
        result = (
            pi
            * A0
            * (jnp.sin(t * pi) - 1j * alpha * jnp.cos(t * pi))
            * integrand(elliptical_path(pi * z))
        )
        return result  # jnp.where(jnp.isfinite(result), result, 0)

    def ytot(t, path=path):
        if path == "complex":
            return yc(t) + y2(t)
        else:
            return y1(t) + y2(t)

    # t0, t1 = 1e-9, 1.0
    t0, t1 = 0.0, 1.0
    if int_method == "romberg":
        return si.romberg(ytot, t0, t1, vec_func=False, show=False, divmax=divmax)
    elif int_method == "romberg_jax":
        if n == 1:
            return romberg(ytot, t0, t1, divmax=divmax)
        else:
            return romberg_vect(n, ytot, t0, t1, divmax=divmax)
    else:
        t = jnp.linspace(t0, t1, n_int)
        return jnp.trapz(ytot(t), t, axis=0)


def trapezcomp(f, a, b, n):
    """
    Composite trapezoidal function integration

    INPUTS:
    f:  the function to integrate
    a:  lower bound of integration
    b:  upper bound
    n:  number of panels to create between ``a`` and ``b``
    """

    # Initialization
    h = (b - a) / n
    x = a

    # Composite rule
    In = f(a)
    for k in range(1, n):
        x = x + h
        In += 2 * f(x)

    return (In + f(b)) * h * 0.5


def romberg(f, a, b, divmax=5, err=False):
    """
    Romberg integration

    INPUTS:
    f:  the function to integrate
    a:  lower bound of integration
    b:  upper bound
    jnp:  number of rows in the Romberg table
    """

    I = jnp.zeros((divmax, divmax), complex)
    error = 0
    for k in range(0, divmax):
        # Composite trapezoidal rule for 2^k panels
        I = ops.index_update(I, ops.index[k, 0], trapezcomp(f, a, b, 2 ** k))

        # Romberg recursive formula
        for j in range(0, k):
            q = (4 ** (j + 1) * I[k, j] - I[k - 1, j]) / (4 ** (j + 1) - 1)
            I = ops.index_update(I, ops.index[k, j + 1], q)
        error = jnp.abs(I[k, -1] - I[k - 1, -2])

    result = I[k, -1]
    if err:
        return result, error
    else:
        return result


def romberg_numpy(f, a, b, eps=1.48e-8, divmax=5):
    """
    Romberg integration

    INPUTS:
    f:  the function to integrate
    a:  lower bound of integration
    b:  upper bound
    jnp:  number of rows in the Romberg table
    """

    I = np.zeros((divmax, divmax), complex)
    error = 0
    for k in range(0, divmax):
        print("k=", k)
        # Composite trapezoidal rule for 2^k panels
        I[k, 0] = trapezcomp(f, a, b, 2 ** k)

        # Romberg recursive formula
        for j in range(0, k):
            q = (4 ** (j + 1) * I[k, j] - I[k - 1, j]) / (4 ** (j + 1) - 1)
            I[k, j + 1] = q
        error = np.abs(I[k, -1] - I[k - 1, -2])
    print(f"error {k} = ", error)
    return I[k, -1]


# import numpy as np
# from jax import lax

# def romberg(f, a, b, eps=1.48e-8, divmax=10):
#     """
#     Romberg integration
#
#     INPUTS:
#     f:  the function to integrate
#     a:  lower bound of integration
#     b:  upper bound
#     divmax:  number of rows in the Romberg table
#     """
#
#     def cond_fun(val):
#         I, k = val
#         error = jnp.where(k > 0, jnp.abs(I[k, -1] - I[k, -2]), jnp.inf)
#         return jnp.where(error > eps,True,False) #and k < divmax
#
#     def body_fun(val):
#         I, k = val
#         # Composite trapezoidal rule for 2^k panelss
#         divmax=1
#         xi = jnp.linspace(a, b, 2 ** divmax +1)
#
#
#
#         def trapezoid(xi,k):
#             h = xi[1]-xi[0]
#             fi = f(xi)
#             s = jnp.sum(fi[1:-1])
#             s = (h / 2) * (fi[0] + fi[-1]) + h * s
#             return s
#
#
#         I = ops.index_update(I, ops.index[k, 0], trapezoid(xi,k))
#         # Romberg recursive formula
#         for j in range(0, k):
#             q = (4 ** (j + 1) * I[k, j] - I[k - 1, j]) / (4 ** (j + 1) - 1)
#             I = ops.index_update(I, ops.index[k, j + 1], q)
#
#         k += 1
#         return I, k
#
#     I = jnp.zeros((divmax, divmax), complex)
#     k = 0
#     init_val =I,k
#     I,k = lax.while_loop(cond_fun, body_fun, init_val)
#
#     return I[k,-1]
#


# while error > eps and k < divmax:
#     # Composite trapezoidal rule for 2^k panels
#     I = ops.index_update(I, ops.index[k, 0], trapezoid(f, a, b, 2 ** k))
#     # Romberg recursive formula
#     for j in range(0, k):
#         q = (4 ** (j + 1) * I[k, j] - I[k - 1, j]) / (4 ** (j + 1) - 1)
#         I = ops.index_update(I, ops.index[k, j + 1], q)
#     if k > 0:
#         error = jnp.abs(I[k, j + 1] - I[k, j])
#     k += 1

# return I[k, j + 1]


def romberg_vect(N, f, *args, **kwargs):
    res = []
    for i in range(N):

        def f_(x):
            return f(x)[i]

        res.append(romberg(f_, *args, **kwargs))
    return jnp.array(res)


if __name__ == "__main__":

    from jax import grad, jit, vmap

    #
    # def f(x):
    #     # return 3 * x ** 2
    #     return jnp.sin(x ** 3)
    #
    # exact = 1
    # bounds = 0, 1
    # npts = 1001
    # Nvec = 3
    #
    # x = jnp.linspace(bounds[0], bounds[1], npts)
    # # param = jnp.linspace(1, 2, 3)
    # # xvec, paramvec = jnp.meshgrid(x, param)
    #
    # def fvec(x):
    #     return jnp.array([x ** 2 + i ** 2 for i in range(Nvec)])
    #
    # def myinteg():
    #     return romberg_vect(Nvec, fvec, *bounds)
    # # N = 10
    # # print(trapezoid(f, *bounds, N))
    # print(romberg(f, *bounds))
    # print(jnp.trapz(f(x), x))
    # print(exact)
    #
    # # num_rombergvect_jax = romberg_vect(Nvec, fvec, *bounds)
    # # print(num_rombergvect_jax)
    # print(myinteg())
    # myintegjit = jit(myinteg)
    # print(myintegjit())
    from coaxtract import CoaxProbe, Layer, Setup, Stack

    h = 430e-6

    def myfunc(eps):

        air = Layer(epsilon=1.0001 - 0.1j, thickness=100e-6)
        substrate = Layer(epsilon=eps - 0.0001j, thickness=h)

        layers = [air, substrate]
        stack = Stack(layers, termination="PEC")
        probe = CoaxProbe()
        setup = Setup(stack, probe)

        freqs = jnp.array([5e9])  # jnp.linspace(5,10,100)*1e9
        k0 = 2 * jnp.pi * freqs / c

        # Npts=10000
        # s=jnp.linspace(0,100,Npts)

        def I(s):
            q = setup.integrand(s, k0)
            return q

        # q = I(s)
        return jnp.sum(path_integral(I, eps, n_int=10)).real

    # import matplotlib.pyplot as plt
    # plt.ion()
    # plt.close("all")
    # plt.plot(s, jnp.abs(q))

    eps = jnp.array([11.2])

    print(myfunc(eps))
    print(grad(myfunc)(eps))

    # def f(s):
    #     return setup.integrand(s, k0)
    #
    # import matplotlib.pyplot as plt
    #
    # plt.ion()
    # plt.close("all")
    # a, b = 0, 110
    #
    # npts = 100000
    #
    # s = jnp.linspace(a, b, npts)
    #
    # fs = f(s)
    #
    # plt.plot(s, jnp.abs(fs))
    # eps = substrate.epsilon
    # s0 = (eps.real) ** 0.5
    # t = jnp.linspace(0, 1, npts)
    #
    # def y1(t):
    #     return 2 * s0 * f(2 * s0 * t)
    #
    # def y2(t):
    #     # return np.where(t==0.0,0.0,1./(t+1e-16))
    #     return jnp.where(
    #         t == 0.0, 0.0, 2 * s0 / (t + 1e-16) ** 2 * f(2 * s0 / (t + 1e-16))
    #     )
    #
    # plt.figure()
    # plt.plot(t, jnp.abs(y1(t)))
    # plt.plot(t, jnp.abs(y2(t)), "--")
    #
    # divmax = 1
    #
    # print("y1")
    # print("--------")
    # print(">>> trapz")
    # print(jnp.trapz(y1(t), t))
    # print(">>> romberg scipy")
    # print(si.romberg(y1, 0, 1, divmax=divmax, vec_func=True))
    # # print(">>> romberg jax")
    # # print(romberg(y1, 0, 1, eps=1e-12, divmax=divmax))
    # print(">>> romberg numpy")
    # print(romberg_numpy(y1, 0, 1, eps=1e-12, divmax=divmax))
    #
    # print("y2")
    # print("--------")
    # print(">>> trapz")
    # print(jnp.trapz(y2(t), t))
    # print(">>> romberg scipy")
    # print(si.romberg(y2, 0, 1, divmax=divmax, vec_func=True))
    # print(">>> romberg jax")
    # print(romberg(y2, 0, 1, eps=1e-12, divmax=divmax))
    #

    # def setup(epsilon):
    #     h = 1
    #     substrate = Layer(epsilon=epsilon, thickness=h)
    #     layers = [substrate]
    #     stack = Stack(layers, termination="PEC")
    #     probe = CoaxProbe()
    #     setup = Setup(stack, probe)
    #
    #     lamb = 0.5
    #     k0 = 2 * jnp.pi / lamb
    #
    #     def f(s):
    #         return setup.integrand(s, k0)
    #
    #     eps = substrate.epsilon
    #     s0 = (eps.real) ** 0.5
    #     t = jnp.linspace(0, 1, npts)
    #
    #     def y1(t):
    #         return 2 * s0 * f(2 * s0 * t)
    #
    #     def y2(t):
    #         return jnp.where(
    #             t == 0.0, 0.0, 2 * s0 / (t + 1e-6) ** 2 * f(2 * s0 / (t + 1e-6))
    #         )
    #
    #     return y2, t
    #     # return jnp.trapz(y2(t), t)
    #
    # #
    #
    # def fun_trapz(epsilon):
    #     y2, t = setup(epsilon)
    #     return jnp.trapz(y2(t), t)
    #
    # #
    #
    # def fun_romberg(epsilon, divmax=3):
    #     y2, t = setup(epsilon)
    #     return romberg(y2, 0, 1, divmax=divmax)
    #
    # #
    # # print(jnp.trapz(y2(t), t))
    # # for divmax in range(1,10):
    # #     print(romberg(y2, 0, 1, divmax=divmax))
    #
    # eps = 12 - 4j
    # I = fun_trapz(eps)
    # print(I)
    # I = fun_romberg(eps)
    # print(I)
    #
    # fun_trapz_jit = jit(fun_trapz)
    # fun_romberg_jit = jit(fun_romberg, static_argnums=(1,))
    # fun_trapz_jit(eps)
    # # fun_trapz_jit(5-7j)
    # fun_romberg_jit(eps)
    # # fun_romberg_jit(5-7j)

    #
    #
    #
    # from jax import jit
    # from functools import partial
    #
    # # @partial(jit, static_argnums=(3,))
    # def trapezoid(f, a, b, N):
    #     h = (b - a) / N
    #     xi = jnp.linspace(a, b, N + 1)
    #     fi = f(xi)
    #     s = jnp.sum(fi[1:-1])
    #     s = (h / 2) * (fi[0] + fi[N]) + h * s
    #     return s
    #
    #
    #
    # def romberg(f, a, b, eps=1.48e-8, divmax=10):
    #     """
    #     Romberg integration
    #
    #     INPUTS:
    #     f:  the function to integrate
    #     a:  lower bound of integration
    #     b:  upper bound
    #     divmax:  number of rows in the Romberg table
    #     """
    #
    #     I = jnp.zeros((divmax, divmax), complex)
    #     for k in range(0, divmax):
    #         # Composite trapezoidal rule for 2^k panels
    #         I = ops.index_update(I, ops.index[k, 0], trapezoid(f, a, b, 2 ** k))
    #         # Romberg recursive formula
    #         for j in range(0, k):
    #             q = (4 ** (j + 1) * I[k, j] - I[k - 1, j]) / (4 ** (j + 1) - 1)
    #             I = ops.index_update(I, ops.index[k, j + 1], q)
    #         if k > 0:
    #             error = jnp.abs(I[k, -1] - I[k, -2])
    #             converged = jnp.where(error < eps, True, False)
    #             if converged:
    #                 break
    #             # return I[k, j+1]
    #         # print(I[k,0:k+1])   # display intermediate results
    #
    #     return I[k, j + 1]
    #
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


"""
Calculation of integrand for a multilayer :cite:p:`Bakhtiari1994`.

"""
# See: Bakhtiari, S., Ganchev, S. I. & Zoughi, R. Analysis of radiation
# from an open-ended coaxial line into stratified dielectrics.
# IEEE Transactions on Microwave Theory and Techniques 42, 1261–1267 (1994)

__all__ = ["Layer", "Stack", "Setup"]

import jax.numpy as jnp

from coaxtract.constants import *
from coaxtract.integ import path_integral
from coaxtract.special import *


class Layer:
    def __init__(self, epsilon, thickness=None, name="layer"):
        self.epsilon = epsilon
        self.thickness = thickness
        self.name = name


class Stack:
    def __init__(self, layers, termination="PEC"):
        if termination not in ["PEC", "infinite"]:
            raise ValueError("termination must be PEC or infinite")
        self.layers = layers
        self.N = len(self.layers)
        self.termination = termination
        self.thicknesses = [l.thickness for l in self.layers]
        self.epsilons = [l.epsilon for l in self.layers]
        if termination == "infinite":
            if self.thicknesses[-1] is not None:
                raise ValueError(
                    "last layer must have None thickness for PEC termination"
                )

    def thickness(self, n):
        thicknesses = [l.thickness for l in self.layers]
        # if self.termination == "infinite":
        #     return sum(self.thicknesses[:n-1])
        # else:
        return sum(self.thicknesses[: (n + 1)])

    def rhoN(self, s, k0):
        if self.termination == "infinite":
            return 0  # jnp.zeros_like(k0)
        else:
            zn = self.thickness(self.N)
            epsN = self.epsilons[-1]
            return jnp.exp(-2 * 1j * k0 * zn * jnp.sqrt(epsN - s ** 2))

    # def beta(self, s, k0, z, rho, eps):
    #     ph = jnp.exp(2 * 1j * k0 * z * jnp.sqrt(eps - s ** 2))
    #     return (1 - rho * ph) / (1 + rho * ph)
    #

    def beta(self, s, k0, z, rho, eps):
        ph = jnp.exp(-2 * 1j * k0 * z * jnp.sqrt(eps - s ** 2))
        return (ph - rho) / (ph + rho + 1e-15)

    def kappa(self, s, eps1, eps2):
        return eps1 / eps2 * jnp.sqrt(eps2 - s ** 2) / jnp.sqrt(eps1 - s ** 2)

    def rho(self, s, k0, z, kappa, beta, eps):
        return (
            (1 - kappa * beta)
            / (1 + kappa * beta)
            * jnp.exp(-2 * 1j * k0 * z * jnp.sqrt(eps - s ** 2))
        )

    def recurrence(self, s, k0):
        n = self.N - 1
        rhos = []
        for i in range(self.N):
            if i == 0:
                rho = self.rhoN(s, k0)
            else:
                z = self.thickness(n)
                beta = self.beta(s, k0, z, rhos[i - 1], self.epsilons[n + 1])
                kappa = self.kappa(s, self.epsilons[n], self.epsilons[n + 1])
                rho = self.rho(s, k0, z, kappa, beta, self.epsilons[n])
            rhos.append(rho)
            n -= 1
        return rhos

    def rho1(self, s, k0):
        return self.recurrence(s, k0)[-1]

    def F(self, s, k0):
        eps = self.epsilons[0]
        rho1 = self.rho1(s, k0)
        result = 1 / jnp.sqrt(eps - s ** 2) * (1 + rho1) / (1 - rho1)
        return result  # jnp.where(jnp.isfinite(result), result, 0)


class Setup:
    def __init__(
        self,
        stack,
        probe,
        bessel="jax",
    ):
        self.stack = stack
        self.probe = probe
        self.bessel = bessel
        if bessel not in ["scipy", "jax"]:
            raise ValueError('bessel must be one of ["scipy", "jax"]')

        def j0app(z):
            return j0_approx(z, N=100)

        self.j0 = j0_scipy if bessel == "scipy" else j0

    def integrand(self, s, k0):
        zout = k0 * s * self.probe.rout
        zin = k0 * s * self.probe.rin
        jout = self.j0(zout)
        jin = self.j0(zin)
        F = self.stack.F(s, k0)
        out = (jout - jin) ** 2 * F / s
        small = F * s * (k0 * (self.probe.rout - self.probe.rin) / 4.0) ** 2
        return jnp.where(s <= 1.0e-5, small, out)

        # return jnp.where(s == 0, 0, out)

    def admittance(
        self,
        freq,
        path="complex",
        alpha=0.5,
        int_method="trapz",
        n_int=700,
        divmax=10,
        iopt=0,
    ):

        if int_method not in ["trapz", "romberg", "romberg_jax"]:
            raise ValueError(
                'bessel must be one of ["trapz", "romberg", "romberg_jax"]'
            )
        k0 = 2 * pi * freq / c
        # eps = jnp.array([jnp.max(jnp.array(self.stack.epsilons).real)])*jnp.ones_like(freq)
        eps = self.stack.epsilons[iopt]

        def _integrand(s):
            return self.integrand(s, k0)

        I = path_integral(
            _integrand,
            eps,
            path,
            alpha,
            int_method,
            n_int,
            divmax,
        )

        y = (
            I
            * eps
            / (self.probe.epsilon ** 0.5 * jnp.log(self.probe.rout / self.probe.rin))
        )
        return y

    def reflection(self, *args, **kwargs):
        y = self.admittance(*args, **kwargs)
        return (1 - y) / (1 + y)


if __name__ == "__main__":

    # Probe
    # Teflon with complex dielectric constant εrc = 2.08 · (1 − 0.0006j).
    # Dimensions for the coax are chosen as a = 0.52 mm and b = 1.2 mm
    from coaxtract import CoaxProbe

    probe = CoaxProbe(epsilon=2.08 * (1 - 0.0006j), rin=0.52e-3, rout=1.2e-3)

    # MUT in Fig. 1 is assumed to have
    # a dielectric constant of εr1 = 2.08 with a thickness of 2.0 mm.

    def myfunc(epsilon):

        layer = Layer(epsilon=epsilon, thickness=2e-3)
        stack = Stack([layer], termination="PEC")
        setup = Setup(stack, probe)
        nconv = 0
        acc = 1e-4
        err = jnp.inf
        n_int = 100

        divmax = 1
        while err > acc:
            y = setup.admittance(
                10e9,
                path="complex",
                alpha=0.5,
                int_method="trapz",
                n_int=n_int,
                divmax=divmax,
            )
            # print(y)
            if nconv > 0:
                err = abs(y - yold)
            nconv += 1
            yold = y
            # print(err)
            n_int *= 2
            divmax += 1

        print(y)
        print(err)
        return y

    from jax import jit

    # fast = jit(myfunc)

    epsr1 = 2.08
    for j in range(-7, 2):
        tand = 10 ** (j)
        epsilon = epsr1 * (1 - 1j * tand)
        myfunc(epsilon)

    # layer = Layer(epsilon=11 - 0.1j, thickness=13e-3)
    # layers = [layer]
    # stack = Stack(layers, termination="PEC")
    # probe = CoaxProbe()
    # setup = Setup(stack, probe)
    # r = setup.reflection(freq=10e9)

    # # from top to bottom
    #
    # air = Layer(epsilon=1 - 0.01j, thickness=1e-2)
    # film = Layer(epsilon=11 - 0.01j, thickness=0.01)
    # substrate = Layer(epsilon=3 - 0.01j, thickness=0.4)
    #
    # # layers = [air, film, substrate]
    #
    # layers = [substrate]
    #
    # stack = Stack(layers, termination="PEC")
    #
    # # # print(stack.tickness(0))
    # #
    # # # rhos = stack.recurrence(1, 1)
    # # # print()
    # # #
    # # s = jnp.linspace(0, 1, 10)
    # # k0 = jnp.linspace(1, 3, 15)
    #
    # s = 0.5
    # k0 = 2
    #
    # from coaxtract import Calibration, CoaxProbe
    #
    # probe = CoaxProbe()
    # cal = Calibration(
    #     probe, epsilon_load=substrate.epsilon, thickness_load=substrate.thickness
    # )
    #
    # #
    # print(">>> one layer")
    # print(stack.F(s, k0))
    # print(cal._integrand_function_load(s, k0))
    #
    # def _integrand_function_tw0_layers(
    #     s, k0, epsilon, thickness, epsilon_substrate, thickness_substrate
    # ):
    #     eps1, eps2 = epsilon, epsilon_substrate
    #     d1, d2 = thickness, thickness_substrate
    #     q1 = jnp.sqrt(eps1 - s ** 2)
    #     q2 = jnp.sqrt(eps2 - s ** 2)
    #     kappa1 = (eps2 * q1) / (eps1 * q2)
    #     t1 = jnp.tan(k0 * d1 * q1)
    #     t2 = jnp.tan(k0 * d2 * q2)
    #     f = (t1 * t2 - kappa1) / (kappa1 * t1 + t2)
    #     return 1j * f / q1
    #
    # air = Layer(epsilon=1 - 0.0j, thickness=1e-62)
    # film = Layer(epsilon=11 - 0.0j, thickness=0.01)
    # substrate = Layer(epsilon=3 - 0.0j, thickness=0.4)
    #
    # layers = [air, substrate]
    #
    # stack = Stack(layers, termination="PEC")
    #
    # i2 = _integrand_function_tw0_layers(
    #     s, k0, air.epsilon, air.thickness, substrate.epsilon, substrate.thickness
    # )
    #
    # print(">>> two layers")
    # print(stack.F(s, k0))
    # print(i2)
    #
    # layer = Layer(epsilon=11 - 0.1j, thickness=13e-3)
    # layers = [layer]
    # stack = Stack(layers, termination="PEC")
    # probe = CoaxProbe()
    # setup = Setup(stack, probe)
    # r = setup.reflection(freq=10e9)
    # print(r)
    # print(jnp.abs(r))
    #
    # film = Layer(epsilon=100 - 0.1j, thickness=10e-6)
    # sublayer = Layer(epsilon=11 - 0.01j, thickness=1e-3)
    # substrate = Layer(epsilon=11 - 0.1j, thickness=0.5e-3)
    # layers = [film, sublayer, substrate]
    # stack = Stack(layers, termination="PEC")
    # probe = CoaxProbe(rin=0.5e-3, rout=3e-3, epsilon=2.1 * (1 - 0.0004j))
    # probe = CoaxProbe()
    # setup = Setup(stack, probe)
    # r = setup.reflection(freq=10e9)
    # print(r)
    # print(jnp.abs(r))
    #
    # cal = Calibration(
    #     probe,
    #     int_method="jax",
    #     epsilon_load=substrate.epsilon,
    #     thickness_load=substrate.thickness,
    # )
    #
    # r = cal.reflection(10e9, "open")
    # print(r)
    #
    # layer = Layer(epsilon=cal.epsilon_open, thickness=None)
    # layers = [layer]
    # stack = Stack(layers, termination="infinite")
    # probe = CoaxProbe()
    # setup = Setup(stack, probe)
    # r = setup.reflection(freq=10e9)
    # print(r)
    #
    # cal = Calibration(
    #     probe,
    #     int_method="trapz",
    #     epsilon_load=substrate.epsilon,
    #     thickness_load=substrate.thickness,
    #     air_gap=0,
    # )
    #
    # r = cal.reflection(10e9, "short")
    # print(r)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


import jax.numpy as jnp
import numpy as np
from jax.config import config

from coaxtract.special import *

config.update("jax_enable_x64", True)


def test_j0():
    z = jnp.linspace(0, 20, 10000)

    assert jnp.allclose(j0(z), j0_scipy(z), atol=5e-6)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = ["load"]

import numpy as npo


def load(filename, Nfreq=801, skiprows=8):
    """Load measurement data.

    Parameters
    ----------
    filename : str
        Name of the file to load.

    Returns
    -------
    freqs: array
        Frequencies (in Hz)

    amplitude: array
        S11 amplitude (in dB)

    phase: array
        S11 phase (in degrees)

    """

    data = npo.loadtxt(filename, skiprows=skiprows, delimiter=",", max_rows=Nfreq)

    freqs, amplitude, phase = data.T[0:3]
    return freqs, amplitude, phase



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = ["CoaxProbe"]

import numpy as np

from .constants import Z0, pi


class CoaxProbe:
    def __init__(
        self, rin=0.994e-3 / 2, rout=5.909e-3 / 2, epsilon=2.25 * (1 - 0.001j)
    ):
        self.rin = rin
        self.rout = rout
        self.epsilon = epsilon
        self.geom_factor = np.log(self.rout / self.rin) / (2 * pi)
        self.impedance = Z0 * self.geom_factor / self.epsilon ** 0.5
